/*
 *               ParaStation
 *
 * Copyright (C) 2003 ParTec AG, Karlsruhe
 * Copyright (C) 2005-2007 ParTec Cluster Competence Center GmbH, Munich
 *
 * $Id: psidnodes.h 4495 2007-01-08 11:14:40Z eicker $
 *
 */
/**
 * @file
 * Functions for handling the various informations about the nodes
 * within the ParaStation daemon
 *
 * $Id: psidnodes.h 4495 2007-01-08 11:14:40Z eicker $
 *
 * @author
 * Norbert Eicker <eicker@par-tec.com>
 *
 */
#ifndef __PSIDNODES_H
#define __PSIDNODES_H

#include <netinet/in.h>

#include "psnodes.h"
#include "pstask.h"

#ifdef __cplusplus
extern "C" {
#if 0
} /* <- just for emacs indentation */
#endif
#endif

/**
 * @brief Initialize the PSIDnodes module.
 *
 * Initialize the PSIDnodes module. It will be prepared to handle @a
 * numNodes nodes.
 *
 * @param numNodes Number of nodes the PSIDnodes module is capable to
 * handle after successful returnd of this function.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_init(PSnodes_ID_t numNodes);

/**
 * @brief Get the number of nodes.
 *
 * Get the actual number of nodes the PSIDnodes module is currently
 * capable to handle after the last call of PSIDnodes_init().
 *
 * @return The actual number of nodes.
 */
PSnodes_ID_t PSIDnodes_getNum(void);

/**
 * @brief Register a new node.
 *
 * Register a new node with ParaStation ID @a id. This node will
 * reside on the host with IP address @a IPaddr. The IP address has to
 * be given in network byteorder.
 *
 * @param id ParaStation ID of the new node.
 *
 * @param addr IP address of the new node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_register(PSnodes_ID_t id, in_addr_t addr);


/**
 * @brief Get the ParaStation ID of a node.
 *
 * Get the ParaStation ID of the node with IP address @a IPaddr. The
 * IP address has to be given in network byteorder.
 *
 * @param addr IP address of the node to lookup.
 *
 * @return If the node was found, the ParaStation ID is returned. Or
 * -1, if an error occured.
 */
PSnodes_ID_t PSIDnodes_lookupHost(in_addr_t addr);

/**
 * @brief Get the IP address of a node.
 *
 * Get the IP address of the node with ParaStation ID @a id. The IP
 * address will be given in network byteorder.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the IP address is returned. Or
 * INADDR_ANY, if an error occured.
 */
in_addr_t PSIDnodes_getAddr(PSnodes_ID_t id);

/**
 * @brief Declare a node to be up.
 *
 * Declare the node with ParaStation ID @a id to be up.
 *
 * @param id ParaStation ID of the node to bring up.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_bringUp(PSnodes_ID_t id);

/**
 * @brief Declare a node to be down.
 *
 * Declare the node with ParaStation ID @a id to be shutdown.
 *
 * @param id ParaStation ID of the node to bring down.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_bringDown(PSnodes_ID_t id);

/**
 * @brief Test if a node is up.
 *
 * Test if the node with ParaStation ID @a id is up.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node is up, 1 is returned. 0 is returned, if an
 * error occured or the node is down.
 */
int PSIDnodes_isUp(PSnodes_ID_t id);


/**
 * @brief Set the protocol version of a node.
 *
 * Set the protocol version the node with ParaStation ID @a id talks
 * to @a version.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param hwType The protocol version of this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setProtocolVersion(PSnodes_ID_t id, int version);

/**
 * @brief Get the protocol version of a node.
 *
 * Get the protocol version the node with ParaStation ID @a id talks.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the protocol version is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_getProtocolVersion(PSnodes_ID_t id);

/**
 * @brief Set the hardware type of a node.
 *
 * Set the hardware type of the node with ParaStation ID @a id to @a hwType.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param hwType The hardware type to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setHWType(PSnodes_ID_t id, int hwType);

/**
 * @brief Get the hardware type of a node.
 *
 * Get the hardware type of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the hardware type is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_getHWType(PSnodes_ID_t id);

/**
 * @brief Set the jobs flag of a node.
 *
 * Set the jobs flag of the node with ParaStation ID @a id to @a runjobs.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param runjobs The runjobs flag to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setRunJobs(PSnodes_ID_t id, int runjobs);

/**
 * @brief Get the jobs flag of a node.
 *
 * Get the jobs flag of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the jobs flag is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_runJobs(PSnodes_ID_t id);

/**
 * @brief Set the starter flag of a node.
 *
 * Set the starter flag of the node with ParaStation ID @a id to @a starter.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param starter The starter flag to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setIsStarter(PSnodes_ID_t id, int starter);

/**
 * @brief Get the starter flag of a node.
 *
 * Get the starter flag of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the starter flag is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_isStarter(PSnodes_ID_t id);

/**
 * @brief Set the extra IP address of a node.
 *
 * Set the extra IP address of the node with ParaStation ID @a id to @a addr.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param addr The extra IP address to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setExtraIP(PSnodes_ID_t id, in_addr_t addr);

/**
 * @brief Get the extra IP address of a node.
 *
 * Get the extra IP address of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the extra IP address is returned. Or
 * INADDR_ANY, if an error occured or the extra IP address was not set.
 */
in_addr_t PSIDnodes_getExtraIP(PSnodes_ID_t id);

/**
 * @brief Set the number of physical CPUs of a node.
 *
 * Set the number of physical CPUs of the node with ParaStation ID @a
 * id to @a numCPU.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param numCPU The number of physical CPUs to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setPhysCPUs(PSnodes_ID_t id, short numCPU);

/**
 * @brief Get the number of physical CPUs of a node.
 *
 * Get the number of physical CPUs of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the number of physical CPUs is
 * returned. Or -1, if an error occured.
 */
short PSIDnodes_getPhysCPUs(PSnodes_ID_t id);

/**
 * @brief Set the number of virtual CPUs of a node.
 *
 * Set the number of virtual CPUs of the node with ParaStation ID @a
 * id to @a numCPU.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param numCPU The number of virtual CPUs to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setVirtCPUs(PSnodes_ID_t id, short numCPU);

/**
 * @brief Get the number of virtual CPUs of a node.
 *
 * Get the number of virtual CPUs of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the number of virtual CPUs is
 * returned. Or -1, if an error occured.
 */
short PSIDnodes_getVirtCPUs(PSnodes_ID_t id);

/**
 * @brief Set the hardware status of a node.
 *
 * Set the hardware status of the node with ParaStation ID @a id to @a
 * hwStatus.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param hwStatus The hardware status to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setHWStatus(PSnodes_ID_t id, int hwStatus);

/**
 * @brief Get the hardware status of a node.
 *
 * Get the hardware status of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the hardware status is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_getHWStatus(PSnodes_ID_t id);

/**
 * Container type to allow the use of PSID_nodes_*GUID() call for
 * both, uid_t and gid_t arguments
 */
typedef union {
    uid_t u;          /**< Use as user ID */
    gid_t g;          /**< Use as group ID */
} PSIDnodes_guid_t;

/** This indicates the type of argument and list to manipulate to the
 * PSID_nodes_*GUID() calls */
typedef enum {
    PSIDNODES_USER,     /**< Manipulate users allowed to use a node*/
    PSIDNODES_GROUP,    /**< Manipulate groups allowed to use a node*/
    PSIDNODES_ADMUSER,  /**< Manipulate node's administrative users */
    PSIDNODES_ADMGROUP, /**< Manipulate node's administrative groups */
} PSIDnodes_gu_t;

/**
 * @brief Set user/group ID to a node.
 *
 * Set the user/group ID of type @a what on the node with ParaStation
 * ID @a id to @a guid.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param what Flag to address the type of ID to set.
 *
 * @param guid The user/group ID to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setGUID(PSnodes_ID_t id,
		      PSIDnodes_gu_t what, PSIDnodes_guid_t guid);

/**
 * @brief Add user/group ID to a node.
 *
 * Add the user/group ID of type @a what with value @a guid to the node
 * with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param what Flag to address the type of ID to add.
 *
 * @param guid The user/group ID to be added to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_addGUID(PSnodes_ID_t id,
		      PSIDnodes_gu_t what, PSIDnodes_guid_t guid);

/**
 * @brief Remove user/group ID from a node.
 *
 * Remove the user/group ID of type @a what with value @a guid from the
 * node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param what Flag to address the type of ID to remove.
 *
 * @param guid The user/group ID to be removed from this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_remGUID(PSnodes_ID_t id,
		      PSIDnodes_gu_t what, PSIDnodes_guid_t guid);

/**
 * @brief Test user/group ID on a node.
 *
 * Test if the user/group ID of type @a what with value @a guid is valid
 * on the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to be tested.
 *
 * @param what Flag to address the type of ID to be tested.
 *
 * @param guid The user/group ID to be tested on this node.
 *
 * @return If the ID to be tested is valid on this node, 1 is
 * returned. Otherwise 0 is given back.
 */
int PSIDnodes_testGUID(PSnodes_ID_t id,
		       PSIDnodes_gu_t what, PSIDnodes_guid_t guid);

/**
 * @brief Send list of user/group IDs.
 *
 * Send an option list of all user/group ID of type @a what to @a
 * dest. Depending on the destination @a dest, the type of GUIDs @a
 * what and the value of the compatibility flag @a compat different
 * types of options are used to actually send.
 *
 * If the compatibility flag @a compat is set, only the first entry of
 * the list of GUIDs is sent using one of the corresponding old (and
 * deprecated) option types.
 *
 * If the destination task is another ParaStation daemon,
 * PSP_OP_SET_[GUID] and one or more PSP_OP_ADD_[GUID] options are
 * used to correctly set the corresponding GUIDs on the other
 * node. Otherwise PSP_OP_[GUID] options are used to send the
 * user-task the corresponding information.
 *
 * @param dest Task ID of the destination task to send to.
 *
 * @param what The type of GUID info to send.
 *
 * @param compat Flag compatibility mode. Only the first GUID in the
 * list will be sent.
 *
 * @return No return value.
 */
void send_GUID_OPTIONS(PStask_ID_t dest, PSIDnodes_gu_t what, int compat);

/**
 * @brief Set the maximum number of processes of a node.
 *
 * Set the maximum number of processes of the node with ParaStation ID
 * @a id to @a procs.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param procs The maximum number of processes to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setProcs(PSnodes_ID_t id, int procs);

/**
 * @brief Get the maximum number of processes of a node.
 *
 * Get the maximum number of processes of the node with ParaStation ID
 * @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the maximum number of processes is
 * returned. Or -1, if an error occured.
 */
int PSIDnodes_getProcs(PSnodes_ID_t id);

/**
 * @brief Set the overbook flag of a node.
 *
 * Set the overbook flag of the node with ParaStation ID @a id to @a
 * overbook.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param overbook The overbook flag to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setOverbook(PSnodes_ID_t id, PSnodes_overbook_t overbook);

/**
 * @brief Get the overbook flag of a node.
 *
 * Get the overbook flag of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the overbook flag is returned. Or
 * -1, if an error occured.
 */
PSnodes_overbook_t PSIDnodes_overbook(PSnodes_ID_t id);

/**
 * @brief Set the exclusive flag of a node.
 *
 * Set the exclusive flag of the node with ParaStation ID @a id to @a
 * exclusive.
 *
 * @param id ParaStation ID of the node to be modified.
 *
 * @param exclusive The exclusive flag to be set to this node.
 *
 * @return On success, 0 is returned or -1, if an error occured.
 */
int PSIDnodes_setExclusive(PSnodes_ID_t id, int exclusive);

/**
 * @brief Get the exclusive flag of a node.
 *
 * Get the exclusive flag of the node with ParaStation ID @a id.
 *
 * @param id ParaStation ID of the node to look up.
 *
 * @return If the node was found, the exclusive flag is returned. Or
 * -1, if an error occured.
 */
int PSIDnodes_exclusive(PSnodes_ID_t id);

#ifdef __cplusplus
}/* extern "C" */
#endif

#endif  /* __PSIDNODES_H */
