/*
 *               ParaStation
 *
 * Copyright (C) 2002-2003 ParTec AG, Karlsruhe
 * Copyright (C) 2005 Cluster Competence Center GmbH, Munich
 *
 * $Id: timer.h 3892 2006-01-09 18:29:54Z eicker $
 *
 */
/**
 * @file
 * ParaStation Timer facility. This is a simple timer multiplexer for
 * applications that need to use independent timers in a transparent
 * way. Within ParaStation this is used by the MCast and RDP modules.
 *
 * $Id: timer.h 3892 2006-01-09 18:29:54Z eicker $
 *
 * @author
 * Norbert Eicker <eicker@par-tec.com>
 *
 */
#ifndef __TIMER_H
#define __TIMER_H

#include <stdint.h>
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#if 0
} /* <- just for emacs indentation */
#endif
#endif

/**
 * @brief Initializes the Timer module.
 *
 * Initialization of the Timer machinery. If any timer is registered,
 * it will be removed.
 *
 * @param logfile File to use for logging. If NULL, syslog(3) is used.
 *
 * @return No return value.
 *
 * @see syslog(3)
 */
void Timer_init(FILE* logfile);

/**
 * @brief Test if the Timer module is initialized.
 *
 * Test if the Timer module is initialized, i.e. if Timer_init() was called
 * before.
 *
 * @return If the Timer module is initialized 1 is returned, 0 otherwise.
 */
int Timer_isInitialized(void);

/**
 * Various message classes for logging. These define the different
 * bits of the debug-mask set via @ref Timer_setDebugMask().
 */ 
typedef enum {
    TIMER_LOG_DUMMY = 0x0000, /**< No non fatal messages yet */
} Timer_log_key_t;

/**
 * @brief Query the debug-mask.
 *
 * Get the debug-mask of the Timer module.
 *
 * @return The actual debug-mask is returned.
 *
 * @see Timer_setDebugMask()
 */
int32_t Timer_getDebugMask(void);

/**
 * @brief Set the debug-mask.
 *
 * Set the debug-mask of the Timer module. @a mask is a bit-wise OR of
 * the different keys defined within @ref Timer_log_key_t. If the
 * respective bit is set within @a mask, the log-messages marked with
 * the corresponding bits are put out to the selected channel
 * (i.e. stderr of syslog() as defined within @ref
 * Timer_init()). Accordingly a @mask of -1 means to put out all
 * messages defined.
 *
 * All messages marked with -1 represent fatal messages that are
 * always put out independently of the choice of @a mask, i.e. even if
 * it is 0.
 *
 * @a mask's default value is 0, i.e. only fatal messages are put out.
 *
 * At the time the Timer module only produces critical errors yet!
 *
 * @param mask The debug-mask to set.
 *
 * @return No return value.
 *
 * @see Timer_getDebugMask(), Timer_log_key_t
 */
void Timer_setDebugMask(int32_t mask);

/**
 * @brief Register a new timer
 *
 * Registration of a new timer. The @a timeoutHandler will be called
 * after @a timeout has elapsed. Afterwards, e.g. for removing using
 * @ref Timer_remove(), the timer will be identified by its unique ID
 * which is returned by this function.
 *
 *
 * @param timeout The amount of time, after which the @a timeoutHandler is
 * called again.
 *
 * @param timeoutHandler If @a timeout has elapsed, this function is called.
 *
 *
 * @return On success, the unique ID of the newly registered timer is
 * returned. This Id is a positive number or 0. On error, -1 is
 * returned.
 */
int Timer_register(struct timeval* timeout, void (*timeoutHandler)(void));

/**
 * @brief Remove a timer
 *
 * Remove a registered timer. The timer will be identified by its
 * corresponding unique ID @a id.
 *
 * @param fd The file-descriptor to identify the timer.
 *
 * @return On success, 0 is returned. On error, -1 is returned.
 */
int Timer_remove(int id);

/**
 * @brief Block a timer
 *
 * Block or unblock a registered timer. The timer will be identified by its
 * corresponding unique ID @a id.
 *
 *
 * @param id The unique ID useed to identify the timer.
 *
 * @param block On 0, the timer will be unblocked. On other values, it will
 *        be blocked.
 *
 *
 * @return If the timer was blocked before, 1 will be returned. If the timer
 * was not blocked, 0 will be returned. If an error occurred, -1 will be
 * returned.
 */
int Timer_block(int id, int block);

/**
 * @brief Handle elapsed timers
 *
 * Within this functions the actual handling of pending timers is
 * done. In order to achieve greater robustness of the code the signal
 * handler within the Timer facility only sets flags marking elapsed
 * timers. The actual work is done within this function.
 *
 * Thus it looks for timers marked to be elapsed and not blocked. If
 * such a timer is found, the corresponding @a timeoutHandler as
 * registered within @ref Timer_register() is called.
 *
 * Since the actual work is done within this function, it is crutial
 * to call it on a regular basis in order to guarantee the handling of
 * the timers. Ideal points are immediately arround central select()
 * calls or somewhere within a main loop of a program.
 *
 * As an example, the Selector facility uses Timers in order to
 * provide certain functionality. Here @ref Timer_handleSignals() is
 * called every time before going into the central select(). Be aware
 * of the fact that this select() is interrupted on a regular basis
 * due to signals sent to the process because of ellapsed timers.
 */
void Timer_handleSignals(void);

#ifdef __cplusplus
}/* extern "C" */
#endif

#endif /* __TIMER_H */
